#include <iostream>
#include <vector>
#include <cassert>

#include <MBCore.hpp>
#include <MBInterface.hpp>
#include <MBAdaptiveKDTree.hpp>
#include <MBRange.hpp>
#include <MBCartVect.hpp>
#include <MBGeomUtil.hpp>

/* Test case demonstrating possible bug in MBAdaptiveKDTree */



/* Ray that triggers bug */
MBCartVect ray_start(-3.276149056544342,  -5,                   3.165678318615075   );
MBCartVect ray_dir  ( 0.2176211564150764, -0.9728521644687909, -0.07873816335775993 );


int main()
{

  /* set up and load */
  MBInterface* mb = new MBCore();
  MBErrorCode rval;
  MBEntityHandle file_set;


  rval = mb->create_meshset( MESHSET_SET, file_set );
  assert( rval == MB_SUCCESS );
  rval = mb->load_file( "tally.cub", &file_set );
  assert( rval == MB_SUCCESS );

  MBRange all_tris;
  rval = mb->get_entities_by_dimension( file_set, 2, all_tris );
  assert( rval == MB_SUCCESS );

  /* Build KD tree (with default parameters) of all triangles */
  MBEntityHandle kdtree_root;
  MBAdaptiveKDTree *kdtree = new MBAdaptiveKDTree( mb );
  kdtree->build_tree( all_tris, kdtree_root );

  /* Test for triangle intersections against ray */
  std::vector< MBEntityHandle > triangles;
  std::vector< double > distances;
  rval = kdtree->ray_intersect_triangles( kdtree_root, 1e-6, ray_dir.array(), ray_start.array(),
					  triangles, distances );
  std::cout << "Triangle intersections found by kd tree test: " << triangles.size() << std::endl;

  /* Iterate over all triangles and perform ray intersection test manually */
  unsigned count = 0;
  for( MBRange::iterator i = all_tris.begin(); i!=all_tris.end(); ++i){
    
    MBEntityHandle triangle = *i;
    std::vector< MBEntityHandle > vertices;
    rval = mb->get_connectivity( &triangle, 1, vertices );
    assert( rval == MB_SUCCESS );

    MBCartVect tri_pts[3];     
    rval = mb->get_coords( &(vertices[0]), 3, tri_pts[0].array() );
    assert( rval == MB_SUCCESS );
    
    double t;
    if( MBGeomUtil::ray_tri_intersect( tri_pts, ray_start, ray_dir, 1e-6, t ) ){
      count++;
    }
  }

  std::cout << "Triangle intersections found by iteration: " << count << std::endl;

  if(triangles.size() == count )
    return 0;
  else
    return 1;
}
